﻿namespace Engine_silk.NET;

using Engine_silk.NET.Textures;
using Nebulix.InputSystem;
using Nebulix.Rendering;
using Silk.NET.Input;
using Silk.NET.Maths;
using Silk.NET.OpenGL;
using Silk.NET.Windowing;
using System;
using System.Drawing;
using System.Numerics;

// Section: https://github.com/dotnet/Silk.NET/tree/main/examples/CSharp/OpenGL%20Tutorials
// Chapter: Continue with 1.3

// Next to do will be cleaning this up by creating a shader class and maybe even an engine class
// IEngine -> IEngine.Create(WindowOptions) -> return OpenGLEngine/VulkanEngine maybe?

public static class Program
{
	private const int Width = 800;
	private const int Height = 600;

	private static IWindow _window;
	private static GL _gl;
	private static Nebulix.Rendering.Shader _shader;
	private static Texture2D _texture;
	private static Camera _cam;
	private static Vector2 _lastMousePosition;
	private static uint _vao, _vbo, _ebo;
	

	public static void Main(string[] args)
	{
		WindowOptions options = WindowOptions.Default with
		{
			Size = new Vector2D<int>(Width, Height),
			Title = "My first Silk.NET application!"
		};
		_window = Window.Create(options);

		_window.Load += OnLoad;
		_window.Update += OnUpdate;
		_window.Render += OnRender;

		_window.Run();
	}

	private static unsafe void OnLoad()
	{
		_gl = _window.CreateOpenGL();
		_gl.ClearColor(Color.CornflowerBlue);
		_gl.Enable(EnableCap.DepthTest);
		//_gl.PolygonMode(GLEnum.FrontAndBack, GLEnum.Fill);

		_cam = new Camera(new(0.0f, 0.0f, 3.0f));
		IInputContext input = _window.CreateInput();
		Input.Init(input);

		for (int i = 0; i < input.Keyboards.Count; i++)
		{
			input.Keyboards[i].KeyDown += OnKeyDown;
		}
		for (int i = 0; i < input.Mice.Count; i++)
		{
			input.Mice[i].MouseMove += OnMouseMove;
			input.Mice[i].Cursor.CursorMode = CursorMode.Disabled;
		}

		_vao = _gl.GenVertexArray();
		_gl.BindVertexArray(_vao);

		float[] vertices =
		{
			// positions          // normals           // texture coords
			-0.5f, -0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  0.0f, 0.0f,
			 0.5f, -0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  1.0f, 0.0f,
			 0.5f,  0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  1.0f, 1.0f,
			 0.5f,  0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  1.0f, 1.0f,
			-0.5f,  0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  0.0f, 1.0f,
			-0.5f, -0.5f, -0.5f,  0.0f,  0.0f, -1.0f,  0.0f, 0.0f,

			-0.5f, -0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   0.0f, 0.0f,
			 0.5f, -0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   1.0f, 0.0f,
			 0.5f,  0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   1.0f, 1.0f,
			 0.5f,  0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   1.0f, 1.0f,
			-0.5f,  0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   0.0f, 1.0f,
			-0.5f, -0.5f,  0.5f,  0.0f,  0.0f, 1.0f,   0.0f, 0.0f,

			-0.5f,  0.5f,  0.5f, -1.0f,  0.0f,  0.0f,  1.0f, 0.0f,
			-0.5f,  0.5f, -0.5f, -1.0f,  0.0f,  0.0f,  1.0f, 1.0f,
			-0.5f, -0.5f, -0.5f, -1.0f,  0.0f,  0.0f,  0.0f, 1.0f,
			-0.5f, -0.5f, -0.5f, -1.0f,  0.0f,  0.0f,  0.0f, 1.0f,
			-0.5f, -0.5f,  0.5f, -1.0f,  0.0f,  0.0f,  0.0f, 0.0f,
			-0.5f,  0.5f,  0.5f, -1.0f,  0.0f,  0.0f,  1.0f, 0.0f,

			 0.5f,  0.5f,  0.5f,  1.0f,  0.0f,  0.0f,  1.0f, 0.0f,
			 0.5f,  0.5f, -0.5f,  1.0f,  0.0f,  0.0f,  1.0f, 1.0f,
			 0.5f, -0.5f, -0.5f,  1.0f,  0.0f,  0.0f,  0.0f, 1.0f,
			 0.5f, -0.5f, -0.5f,  1.0f,  0.0f,  0.0f,  0.0f, 1.0f,
			 0.5f, -0.5f,  0.5f,  1.0f,  0.0f,  0.0f,  0.0f, 0.0f,
			 0.5f,  0.5f,  0.5f,  1.0f,  0.0f,  0.0f,  1.0f, 0.0f,

			-0.5f, -0.5f, -0.5f,  0.0f, -1.0f,  0.0f,  0.0f, 1.0f,
			 0.5f, -0.5f, -0.5f,  0.0f, -1.0f,  0.0f,  1.0f, 1.0f,
			 0.5f, -0.5f,  0.5f,  0.0f, -1.0f,  0.0f,  1.0f, 0.0f,
			 0.5f, -0.5f,  0.5f,  0.0f, -1.0f,  0.0f,  1.0f, 0.0f,
			-0.5f, -0.5f,  0.5f,  0.0f, -1.0f,  0.0f,  0.0f, 0.0f,
			-0.5f, -0.5f, -0.5f,  0.0f, -1.0f,  0.0f,  0.0f, 1.0f,

			-0.5f,  0.5f, -0.5f,  0.0f,  1.0f,  0.0f,  0.0f, 1.0f,
			 0.5f,  0.5f, -0.5f,  0.0f,  1.0f,  0.0f,  1.0f, 1.0f,
			 0.5f,  0.5f,  0.5f,  0.0f,  1.0f,  0.0f,  1.0f, 0.0f,
			 0.5f,  0.5f,  0.5f,  0.0f,  1.0f,  0.0f,  1.0f, 0.0f,
			-0.5f,  0.5f,  0.5f,  0.0f,  1.0f,  0.0f,  0.0f, 0.0f,
			-0.5f,  0.5f, -0.5f,  0.0f,  1.0f,  0.0f,  0.0f, 1.0f
		};
		ReadOnlySpan<float> verticesData = new(vertices);
		//uint[] indices =
		//{
		//	0u, 1u, 3u,
		//	1u, 2u, 3u,

		//	4u, 5u, 7u,
		//	5u, 6u, 7u,

		//	3u, 2u, 7u,
		//	7u, 2u, 6u,

		//	0u, 4u, 5u,
		//	0u, 5u, 1u,

		//	0u, 3u, 7u,
		//	0u, 7u, 4u,

		//	1u, 6u, 5u,
		//	1u, 2u, 6u
		//};

		_vbo = _gl.GenBuffer();
		_gl.BindBuffer(BufferTargetARB.ArrayBuffer, _vbo);
		_gl.BufferData(BufferTargetARB.ArrayBuffer, (nuint)(verticesData.Length * sizeof(float)), verticesData, BufferUsageARB.StaticDraw);

		//_ebo = _gl.GenBuffer();
		//_gl.BindBuffer(BufferTargetARB.ElementArrayBuffer, _ebo);
		//fixed (uint* buffer = indices)
		//	_gl.BufferData(BufferTargetARB.ElementArrayBuffer, (nuint)(indices.Length * sizeof(uint)), buffer, BufferUsageARB.StaticDraw);

		_shader = new Nebulix.Rendering.Shader(_gl, "shader.vert", "shader.frag");

		_gl.EnableVertexAttribArray(0);
		_gl.VertexAttribPointer(0, 3, VertexAttribPointerType.Float, false, 8 * sizeof(float), (void*)0);
		_gl.EnableVertexAttribArray(1);
		_gl.VertexAttribPointer(1, 3, VertexAttribPointerType.Float, false, 8 * sizeof(float), (void*)(3 * sizeof(float)));
		_gl.EnableVertexAttribArray(2);
		_gl.VertexAttribPointer(2, 2, VertexAttribPointerType.Float, false, 8 * sizeof(float), (void*)(6 * sizeof(float)));

		_texture = new Texture2D(_gl, "images/container.png", ImageFormat.RGBA);
	}

	private static void OnUpdate(double deltaTime)
	{
		if(Input.AnyKeyPressed())
		{
			const float movementSpeedMultiplier = 2;
			if(Input.IsKeyPressed(Key.ShiftLeft))
				_cam.MovementSpeed *= movementSpeedMultiplier;

			if (Input.IsKeyPressed(Key.W))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Forward);
			if(Input.IsKeyPressed(Key.S))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Backward);
			if (Input.IsKeyPressed(Key.A))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Left);
			if(Input.IsKeyPressed(Key.D))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Right);
			if (Input.IsKeyPressed(Key.Space))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Up);
			if(Input.IsKeyPressed(Key.ControlLeft))
				_cam.ProcessKeyboard((float)deltaTime, MovementDirection.Down);

			if (Input.IsKeyPressed(Key.ShiftLeft))
				_cam.MovementSpeed /= movementSpeedMultiplier;
		}
	}

	private static void OnRender(double deltaTime)
	{
		_gl.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

		var difference = (float)(_window.Time * 10);
		var modelMatrix = Matrix4x4.CreateRotationY(Maths.Convert.ToRadians(difference)) * Matrix4x4.CreateRotationX(Maths.Convert.ToRadians(difference));
		var viewMatrix = _cam.ViewMatrix;
		var projectionMatrix = Matrix4X4.CreatePerspectiveFieldOfView(Maths.Convert.ToRadians(_cam.Fov), Width / (float)Height, 0.1f, 100.0f);

		_gl.BindVertexArray(_vao);

		_shader.Use();
		_shader.SetMatrix("modelMatrix", modelMatrix);
		_shader.SetMatrix("projectionMatrix", projectionMatrix);
		_shader.SetMatrix("viewMatrix", viewMatrix);

		_shader.SetInt("tex", 0);
		_texture.Bind();

		_gl.DrawArrays(PrimitiveType.Triangles, 0, 36);
	}

	private static void OnKeyDown(IKeyboard keyboard, Key key, int keyCode)
	{
		if (key == Key.Escape)
		{
			_window.Close();
		}
	}

	private static void OnMouseMove(IMouse mouse, Vector2 position)
	{
		if(_lastMousePosition == default)
		{
			_lastMousePosition = position;
			return;
		}

		float xOffset = position.X - _lastMousePosition.X;
		float yOffset = _lastMousePosition.Y - position.Y;
		
		_lastMousePosition = position;
		_cam.ProcessMouseMovement(xOffset, yOffset);
	}

}
