﻿using Nebulix;
using Nebulix.Rendering;
using Silk.NET.Maths;
using Silk.NET.OpenGL;

namespace Engine_silk.NET
{
	// also maybe make an ISphere and call this class "CubeSphere"?
	public class Sphere(uint resolution)
	{
		private readonly Face[] sphereFaces = new Face[6];
		public void CreateSphere()
		{
			Vector3D<float>[] directions =
			{
				Vector3D<float>.UnitZ, -Vector3D<float>.UnitZ, 
				Vector3D<float>.UnitY, -Vector3D<float>.UnitY,
				Vector3D<float>.UnitX, -Vector3D<float>.UnitX
			};

			for (int i = 0; i < sphereFaces.Length; i++)
			{
				// TODO: Refactor so that a face gets a mesh it can only change the contents of (should improve performance)
				sphereFaces[i] = new Face(directions[i], resolution);
			}
		}
		
		public void RenderSphere(GL gl) // Will not be needed  
		{
			// Use default shader etc. to render the sphere
			for (int i = 0; i < 6; i++)
			{
				Mesh m = sphereFaces[i].GetMesh();
				m.Use(gl);
			}
		}
	}

	internal record struct Face
	{
		private readonly Vector3D<float> _localX;
		private readonly Vector3D<float> _localY;

		private readonly Vector3D<float> _localUp;
		private readonly uint _resolution;
		
		public Face(Vector3D<float> localUp, uint resolution)
		{
			_localX = new(localUp.Y, localUp.Z, localUp.X);
			_localY = Vector3D.Cross(localUp, _localX);
			_localUp = localUp;
			_resolution = resolution;
		}

		internal Mesh GetMesh()
		{
			var vertices = new Vector3D<float>[_resolution * _resolution];
			// _resolution - 1 because the vertices index starts at 0 
			// * 6 because each triangle needs 3 points and each small quad has 2 triangles 3*2 = 6
			nuint[] indices = new nuint[(_resolution - 1) * (_resolution - 1) * 6];
			int triangleIndex = 0;

			uint i;
			for (uint y = 0; y < _resolution; y++)
			{
				for (uint x = 0; x < _resolution; x++)
				{
					i = x + y * _resolution;
					Vector2D<float> percent = new Vector2D<float>(x, y) / (_resolution - 1);
					// place vertex on correct position of the plane to easily calculate indices
					Vector3D<float> vertexPosition = _localUp + (percent.X - 0.5f) * 2 * _localX + (percent.Y - 0.5f) * 2 * _localY;
					vertices[i] = Vector3D.Normalize(vertexPosition); // normalise vertex position to get it to be "on the sphere" and not "on the plane"

					if (x != _resolution - 1 && y != _resolution - 1) // we didn't reach the bottom right point yet
					{
						indices[triangleIndex] = i;
						indices[triangleIndex + 1] = i + _resolution + 1;
						indices[triangleIndex + 2] = i + _resolution;
						
						indices[triangleIndex + 3] = i;
						indices[triangleIndex + 4] = i + 1;
						indices[triangleIndex + 5] = i + _resolution + 1;

						triangleIndex += 6; // doing it this way is faster than doing indices[triangleIndex++] (takes about double the time)
					}
				}
			}

			Mesh m = new Mesh();
			m.Vertices = vertices.ExtractComponents();
			m.Indices = indices;

			return m;
		}
	}
}
